/* -----------------------------------------------------------------------

   Company: Netcope Technologies a.s.


   Project: Netcope Drivers



   -------------------------------------------------------------------------


   (c) Copyright ((YEARS)) Netcope Technologies a.s.
   All rights reserved.

   Please review the terms of the license agreement before using this
   file. If you are not an authorized user, please destroy this
   source code file and notify Netcope Technologies a.s. immediately
   that you inadvertently received an unauthorized copy.

   ------------------------------------------------------------------------
*/
/*
 * Xilinx XVC PCIe Driver
 * Copyright (C) 2017 Xilinx Corporation
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include <linux/types.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/module.h>

#include "xvc_pcie_driver_base.h"
#include "xvc_pcie_driver.h"
#include "xvc_pcie_user_config.h"

#ifdef INCLUDE_CUSTOM_MMCONFIG
#include "mmconfig.h"
#endif

struct xil_xvc_char {
	struct pci_dev *pci_dev;
	struct xvc_algo_t xvc_algo;
};

static struct class *xvc_dev_class;

static const size_t CFG_DEVICES_MAX = 10;
static struct xil_xvc_char xil_xvc_devices[10];

static int xvc_ioc_dev_region_status = -1;
static dev_t xvc_ioc_dev_region;
static struct cdev xvc_char_ioc_dev;

static dev_t get_device_number(dev_t region, int offset) {
	return MKDEV(MAJOR(region), MINOR(region) + offset);
}

static void xil_xvc_cleanup(void) {
	printk(KERN_INFO LOG_PREFIX "Cleaning up resources...\n");

	if (!IS_ERR(xvc_dev_class)) {
		class_unregister(xvc_dev_class);
		class_destroy(xvc_dev_class);
	}

	if (xvc_char_ioc_dev.owner != NULL) {
		cdev_del(&xvc_char_ioc_dev);
	}

	if (xvc_ioc_dev_region_status == 0) {
		unregister_chrdev_region(xvc_ioc_dev_region, CFG_DEVICES_MAX);
	}
}

static long setup_xvc_algo(struct xil_xvc_char *xvc_char) {
	int status = -EINVAL;
	int bar_index = user_config.bar_info.bar_index;

	// initialize algo selection if not already selected
	if (!xvc_char->xvc_algo.type) {

		struct xvc_algo_t *algo = &xvc_char->xvc_algo;

		// try to find the cfg capability if CONFIG or AUTO was specified by user
		if (user_config.config_space == CONFIG || user_config.config_space == AUTO) {
			status = xil_xvc_get_offset(xvc_char->pci_dev, &algo->offset.cfg);
			if (status == 0) {
				// found the capability so use CFG space;
				algo->type = XVC_ALGO_CFG;
				return status;
			}
		}
		// try to use BAR space if selected, or fall through for AUTO
		if (user_config.config_space == BAR || user_config.config_space == AUTO) {
			resource_size_t bar_start;
			resource_size_t bar_len;
			resource_size_t map_len;

			if ((!pci_resource_flags(xvc_char->pci_dev, bar_index)) & IORESOURCE_MEM) {
				printk(KERN_ERR LOG_PREFIX "Incorrect BAR configuration\n");
				return -ENODEV;
			}

			bar_start = pci_resource_start(xvc_char->pci_dev, bar_index);
			bar_len = pci_resource_len(xvc_char->pci_dev, bar_index);
			map_len = bar_len;

			if (!bar_len) {
				printk(KERN_WARNING LOG_PREFIX "BAR #%d is not present.\n", bar_index);
			} else {
				// add user specified BAR offset to base address of mapping
				algo->offset.bar = pci_iomap(xvc_char->pci_dev, bar_index, map_len) + user_config.bar_info.bar_offset;
				if (!(algo->offset.bar)) {
					printk(KERN_ERR LOG_PREFIX "Could not map BAR %d\n", bar_index);
					return -ENODEV;
				} else {
					algo->type = XVC_ALGO_BAR;
					printk(KERN_INFO LOG_PREFIX "BAR%d at 0x%llx mapped at 0x%p, length=%llu(/%llu)\n", bar_index,
							(u64)bar_start, algo->offset.bar, (u64)map_len, (u64)bar_len);
					status = 0;
				}
			}
		}
	}

	return status;
}

long char_ctrl_ioctl(struct file *file_p, unsigned int cmd, unsigned long arg) {
	int device_index = iminor(file_p->f_path.dentry->d_inode) - MINOR(xvc_ioc_dev_region);
	int status = 0;
	unsigned long pci_config_lock_flags = 0;

	if (xil_xvc_devices[device_index].pci_dev == NULL) {
		printk(KERN_ERR LOG_PREFIX "Could not find device_index %d\n", device_index);
		return -EFAULT;
	}

	spin_lock_irqsave(&file_p->f_path.dentry->d_inode->i_lock, pci_config_lock_flags);

	switch (cmd)
	{
	case XDMA_IOCXVC:
		status = xil_xvc_ioctl(xil_xvc_devices[device_index].pci_dev, &xil_xvc_devices[device_index].xvc_algo, (void __user *)arg);
		break;
	default:
		status = -ENOIOCTLCMD;
		break;
	}

// mmiowb has been removed since Linux 5.2, and on platforms where it's needed,
// it's hidden inside spin_unlock(), see https://lkml.org/lkml/2019/5/6/1048
#ifdef mmiowb
	mmiowb();
#endif
	spin_unlock_irqrestore(&file_p->f_path.dentry->d_inode->i_lock, pci_config_lock_flags);

	return status;
}

static struct file_operations xil_xvc_ioc_ops = {
	.owner = THIS_MODULE,
	.unlocked_ioctl = char_ctrl_ioctl
};


// -----------------------
// -----------------------
// PCI initialization code
// -----------------------
// -----------------------
int nfb_xvc_probe(struct nfb_device *nfb, void **priv)
{
	dev_t ioc_device_number;
	char ioc_device_name[25];
	struct device *xvc_ioc_device;
	struct pci_dev *pdev = nfb->pci;

	int status = 0;

	int i = 0;
	for (i = 0; i < CFG_DEVICES_MAX; i++) {
		if (xil_xvc_devices[i].pci_dev == NULL) {
				// Add device to table
			xil_xvc_devices[i].pci_dev = pdev;
			memset(&xil_xvc_devices[i].xvc_algo, 0, sizeof(struct xvc_algo_t));
				status = setup_xvc_algo(xil_xvc_devices + i);
			if (status != 0) {
				printk(KERN_ERR LOG_PREFIX "Failed to setup xvc capability\n");
				xil_xvc_devices[i].pci_dev = NULL;
				return status;
			}
				// Character driver file
			ioc_device_number = get_device_number(xvc_ioc_dev_region, i);
				// Make device character file name with right number
			sprintf(ioc_device_name, "xil_xvc/cfg_ioc%d", i);
				xvc_ioc_device = device_create(xvc_dev_class, NULL, ioc_device_number, NULL, ioc_device_name);
			if (IS_ERR(xvc_ioc_device)) {
				printk(KERN_WARNING LOG_PREFIX "Failed to create the device %d %s\n", i, ioc_device_name);
			} else {
				printk(KERN_INFO LOG_PREFIX "Created device %d %s, reg offset (0x%lX)\n", i, ioc_device_name,
						xil_xvc_devices[i].xvc_algo.offset.cfg);
			}
				break;
		}
	}
	return status;
}

void nfb_xvc_remove(struct nfb_device *nfb, void *priv)
{
	int i = 0;
	struct pci_dev *pdev = nfb->pci;

	for (i = 0; i < CFG_DEVICES_MAX; i++) {
		if (pdev != NULL && xil_xvc_devices[i].pci_dev == pdev) {
			if (xil_xvc_devices[i].xvc_algo.type == XVC_ALGO_BAR) {
				pci_iounmap(pdev, xil_xvc_devices[i].xvc_algo.offset.bar);
				printk(KERN_INFO LOG_PREFIX "Unmapping BAR%d.\n", user_config.bar_info.bar_index);
			}

			xil_xvc_devices[i].pci_dev = NULL;

			device_destroy(xvc_dev_class, get_device_number(xvc_ioc_dev_region, i));

			printk(KERN_INFO LOG_PREFIX "Device removal found at %d\n", i);
			break;
		}
	}
}

static struct nfb_driver_ops nfb_xvc_driver_ops = {
	.attach = nfb_xvc_probe,
	.detach = nfb_xvc_remove,
};

// --------------------------
// --------------------------
// Driver initialization code
// --------------------------
// --------------------------

int __init xil_xvc_init()
{
	printk(KERN_INFO LOG_PREFIX "Starting...\n");

	// Register the character packet device major and minor numbers
	xvc_ioc_dev_region_status = alloc_chrdev_region(&xvc_ioc_dev_region, 0,
			CFG_DEVICES_MAX, "xilinx_xvc_pci_ioc_driver_region");
	if (xvc_ioc_dev_region_status != 0) {
		xil_xvc_cleanup();
		return xvc_ioc_dev_region_status;
	}

	// Add the character device, no actual files yet
	cdev_init(&xvc_char_ioc_dev, &xil_xvc_ioc_ops);
	xvc_char_ioc_dev.owner = THIS_MODULE;
	cdev_add(&xvc_char_ioc_dev, xvc_ioc_dev_region, CFG_DEVICES_MAX);

	// Register the character device class for the actual files
#ifdef CONFIG_CLASS_CREATE_HAVE_ONE_PARAMETER
	xvc_dev_class = class_create("xil_xvc_class");
#else
	xvc_dev_class = class_create(THIS_MODULE, "xil_xvc_class");
#endif
	if (IS_ERR(xvc_dev_class)) {
		xil_xvc_cleanup();
		return PTR_ERR(xvc_dev_class);
	}

	// init device list
	memset(xil_xvc_devices, 0, sizeof(xil_xvc_devices));

#ifdef INCLUDE_CUSTOM_MMCONFIG
	// Comment this out on 32 bit systems
	pci_mmcfg_driver_init();
#endif

	nfb_driver_register(nfb_xvc_driver_ops);

	return 0;
}


void __exit xil_xvc_exit()
{
	nfb_driver_unregister(nfb_xvc_driver_ops);
#ifdef INCLUDE_CUSTOM_MMCONFIG
	// Comment this out on 32 bit systems
	pci_mmcfg_driver_exit();
#endif

	xil_xvc_cleanup();
}

MODULE_VERSION(PACKAGE_VERSION);
MODULE_LICENSE("GPL");
module_init(xil_xvc_init);
module_exit(xil_xvc_exit);
